from __future__ import annotations

import re
from os.path import dirname, join

from winval import logger
from winval.parsed_workflow_inputs import ParsedWorkflowInputs
from winval.workflow_var import WorkflowVar, WorkflowVarType


class WdlParser:

    def __init__(self, wdl_file):
        self.wdl_file = wdl_file
        self.structs: dict[str, dict[str, WorkflowVar]] = {}
        self.workflow_vars: dict[str, WorkflowVar] = {}
        self.imported_wdls = set()
        self.workflow_name = None

    def parse_workflow_variables(self) -> ParsedWorkflowInputs:
        wdl_code_lines = self.__read_wdl_code_lines(self.wdl_file)
        self._parse_workflow_name(wdl_code_lines)
        self.__parse_structs(self.wdl_file)
        self._parse_variables_from_lines(wdl_code_lines)
        return ParsedWorkflowInputs(self.workflow_vars, self.structs)

    def __read_wdl_code_lines(self, wdl_file: str):
        with open(wdl_file) as wf:
            wdl_lines = wf.readlines()
            wdl_code_lines = self.get_code_lines(wdl_lines)
        return wdl_code_lines

    @staticmethod
    def get_code_lines(wdl_lines):
        wdl_code_lines = [strip_comments(line) for line in wdl_lines if is_code_line(line)]
        return wdl_code_lines

    def _parse_workflow_name(self, wdl_lines):
        for line in wdl_lines:
            line = line.strip()
            tokens = re.split(r'\s+', line)
            if tokens[0] == 'workflow':
                self.workflow_name = re.match(r'\w*', tokens[1]).group(0)
                logger.debug(f'workflow: {self.workflow_name}')
                return

    def _parse_variables_from_lines(self, wdl_lines) -> None:
        # For extra safety, in case someone uses this protected function from outside
        wdl_lines = self.get_code_lines(wdl_lines)
        for line_num, line in enumerate(wdl_lines):
            line = line.strip()
            tokens = re.split(r'\s+', line)

            if re.match('input{?', tokens[0]):
                if self.workflow_name is None:
                    raise ValueError('Undefined workflow name')
                workflow_vars = self.__parse_variables_from_block(wdl_lines[(line_num + 1):])
                self.workflow_vars.update(workflow_vars)
                return
        raise ValueError('Could not parse workflow args')

    def __parse_variables_from_block(self, wdl_lines_after_block_start: list) -> dict[str, WorkflowVar]:
        """
        Internal parsing function which is used both for workflow inputs variables and structs fields
        :param wdl_lines_after_block_start:
        :return dictionary of workflow variables:
        """
        workflow_vars = {}
        curly_brackets_scope = 1
        straight_brackets_scope = 0

        while len(wdl_lines_after_block_start) > 0:
            # end of inputs block
            if curly_brackets_scope <= 0:
                break
            line = wdl_lines_after_block_start.pop(0)
            curly_brackets_scope, straight_brackets_scope = \
                self.__update_scope(curly_brackets_scope, line, straight_brackets_scope)

            # inside internal brackets, meaning multiline complex variable default init
            while curly_brackets_scope > 1 or straight_brackets_scope > 0:
                next_line = wdl_lines_after_block_start.pop(0)
                line += next_line
                curly_brackets_scope, straight_brackets_scope = \
                    self.__update_scope(curly_brackets_scope, next_line, straight_brackets_scope)

            line = line.strip()
            if len(line) == 0 or line.startswith('}'):
                continue

            # parse single-line default value (if default is multiline, will ignore all but first
            equal_split = line.split('=')
            if len(equal_split) == 0 or equal_split == ['']:
                continue
            var_str = equal_split[0]
            if len(equal_split) == 2:  # has default value
                default_value_str = equal_split[1].strip()
            else:
                default_value_str = None

            var_tokens = [token for token in re.split('[ ]', var_str) if token != '']
            var_name = var_tokens[-1]
            var_type = ''.join(var_tokens[:-1])
            is_optional = '?' in var_type
            var_type = var_type.replace('?', '')
            # ignore is_non_empty qualifier
            var_type = var_type.replace('+', '')

            if default_value_str is not None:
                is_optional = True

            # If struct is used, add all of its fields as variables to workflow
            if var_type in self.structs:
                workflow_var_type = WorkflowVarType.STRUCT
                struct_type = var_type
            else:
                workflow_var_type = WorkflowVarType.from_string(var_type)
                struct_type = None

            workflow_vars[var_name] = WorkflowVar(var_name,
                                                  workflow_var_type,
                                                  is_optional,
                                                  default_value_str,
                                                  self.workflow_name,
                                                  struct_type)
        return workflow_vars

    def __update_scope(self, curly_brackets_scope, line, straight_brackets_scope):
        curly_brackets_scope -= line.count('}')
        curly_brackets_scope += line.count('{')
        straight_brackets_scope -= line.count(']')
        straight_brackets_scope += line.count('[')
        return curly_brackets_scope, straight_brackets_scope

    def __parse_structs(self, wdl_file) -> None:
        self.imported_wdls.update({wdl_file})
        wdl_lines = self.__read_wdl_code_lines(wdl_file)
        self._parse_structs_iter(wdl_lines)
        for line in wdl_lines:
            line = line.strip()
            tokens = re.split(r'\s+', line)
            # stop parsing at workflow/task definition
            if tokens[0] in ['workflow', 'task']:
                return
            if tokens[0] == 'import':
                import_wl = tokens[1].replace('"', '').replace("'", '')
                import_path = join(dirname(wdl_file), import_wl)
                if import_path not in self.imported_wdls:
                    logger.debug(f'importing structs from {import_path}')
                    self.__parse_structs(import_path)

    def _parse_structs_iter(self, wdl_lines: list[str]) -> None:
        # For extra safety, in case someone uses this protected function from outside
        wdl_lines = self.get_code_lines(wdl_lines)
        wdl_text = '\n'.join(wdl_lines)
        struct_paragraphs: list[str] = \
            re.findall(r'struct\s+\w+\s*{\s*\n*[\[\]\w+\s*?\n]*}', wdl_text)

        for struct_paragraph in struct_paragraphs:
            struct_lines = re.split(r'[\n{}]', struct_paragraph)
            tokens = re.split(r'\s+', struct_lines[0])
            struct_name = tokens[1]
            logger.debug(f'\t{struct_name}')
            struct_variables = self.__parse_variables_from_block(struct_lines[1:])
            self.structs[struct_name] = struct_variables


def is_code_line(line: str) -> bool:
    line = line.strip()
    tokens = re.split(r'\s+', line)
    is_comment = tokens[0].startswith('#')
    is_empty = len(tokens) == 0 or tokens == ['']
    return not (is_comment or is_empty)


def strip_comments(line: str) -> str:
    return re.split('#', line)[0].strip()



# Generated from winval.g4 by ANTLR 4.10.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,40,99,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,1,0,1,0,1,0,1,1,1,1,1,
        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,3,1,38,8,1,1,1,1,1,1,1,1,1,1,1,1,1,
        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,5,1,58,8,1,10,1,
        12,1,61,9,1,1,2,1,2,1,3,1,3,1,3,5,3,68,8,3,10,3,12,3,71,9,3,1,3,
        1,3,1,3,5,3,76,8,3,10,3,12,3,79,9,3,1,3,1,3,1,3,5,3,84,8,3,10,3,
        12,3,87,9,3,1,3,1,3,1,3,5,3,92,8,3,10,3,12,3,95,9,3,3,3,97,8,3,1,
        3,0,1,2,4,0,2,4,6,0,6,1,0,7,13,1,0,14,19,1,0,21,22,1,0,23,28,1,0,
        29,30,1,0,31,32,117,0,8,1,0,0,0,2,37,1,0,0,0,4,62,1,0,0,0,6,96,1,
        0,0,0,8,9,3,2,1,0,9,10,5,0,0,1,10,1,1,0,0,0,11,12,6,1,-1,0,12,38,
        5,36,0,0,13,38,5,37,0,0,14,38,5,39,0,0,15,38,5,34,0,0,16,38,5,38,
        0,0,17,38,3,4,2,0,18,19,3,4,2,0,19,20,5,1,0,0,20,21,3,2,1,0,21,22,
        5,2,0,0,22,38,1,0,0,0,23,24,5,1,0,0,24,25,3,6,3,0,25,26,5,2,0,0,
        26,38,1,0,0,0,27,28,5,3,0,0,28,29,3,6,3,0,29,30,5,4,0,0,30,38,1,
        0,0,0,31,32,5,5,0,0,32,33,3,2,1,0,33,34,5,6,0,0,34,38,1,0,0,0,35,
        36,7,0,0,0,36,38,3,2,1,7,37,11,1,0,0,0,37,13,1,0,0,0,37,14,1,0,0,
        0,37,15,1,0,0,0,37,16,1,0,0,0,37,17,1,0,0,0,37,18,1,0,0,0,37,23,
        1,0,0,0,37,27,1,0,0,0,37,31,1,0,0,0,37,35,1,0,0,0,38,59,1,0,0,0,
        39,40,10,6,0,0,40,41,7,1,0,0,41,58,3,2,1,7,42,43,10,5,0,0,43,44,
        5,20,0,0,44,58,3,2,1,6,45,46,10,4,0,0,46,47,7,2,0,0,47,58,3,2,1,
        5,48,49,10,3,0,0,49,50,7,3,0,0,50,58,3,2,1,4,51,52,10,2,0,0,52,53,
        7,4,0,0,53,58,3,2,1,3,54,55,10,1,0,0,55,56,7,5,0,0,56,58,3,2,1,2,
        57,39,1,0,0,0,57,42,1,0,0,0,57,45,1,0,0,0,57,48,1,0,0,0,57,51,1,
        0,0,0,57,54,1,0,0,0,58,61,1,0,0,0,59,57,1,0,0,0,59,60,1,0,0,0,60,
        3,1,0,0,0,61,59,1,0,0,0,62,63,5,35,0,0,63,5,1,0,0,0,64,69,5,36,0,
        0,65,66,5,33,0,0,66,68,5,36,0,0,67,65,1,0,0,0,68,71,1,0,0,0,69,67,
        1,0,0,0,69,70,1,0,0,0,70,97,1,0,0,0,71,69,1,0,0,0,72,77,5,37,0,0,
        73,74,5,33,0,0,74,76,5,37,0,0,75,73,1,0,0,0,76,79,1,0,0,0,77,75,
        1,0,0,0,77,78,1,0,0,0,78,97,1,0,0,0,79,77,1,0,0,0,80,85,5,39,0,0,
        81,82,5,33,0,0,82,84,5,39,0,0,83,81,1,0,0,0,84,87,1,0,0,0,85,83,
        1,0,0,0,85,86,1,0,0,0,86,97,1,0,0,0,87,85,1,0,0,0,88,93,5,35,0,0,
        89,90,5,33,0,0,90,92,5,35,0,0,91,89,1,0,0,0,92,95,1,0,0,0,93,91,
        1,0,0,0,93,94,1,0,0,0,94,97,1,0,0,0,95,93,1,0,0,0,96,64,1,0,0,0,
        96,72,1,0,0,0,96,80,1,0,0,0,96,88,1,0,0,0,97,7,1,0,0,0,8,37,57,59,
        69,77,85,93,96
    ]

class winvalParser ( Parser ):

    grammarFileName = "winval.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'['", "']'", "'{'", "'}'", "'('", "')'", 
                     "'defined'", "'len'", "'prefix'", "'suffix'", "'basename'", 
                     "'splitext'", "'not'", "'+'", "'-'", "'*'", "'/'", 
                     "'**'", "'%'", "'in'", "'|'", "'&'", "'>'", "'>='", 
                     "'=='", "'<='", "'<'", "'!='", "'or'", "'and'", "'->'", 
                     "'<->'", "','" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "BOOL", "VARIABLE", "INT", 
                      "FLOAT", "EMPTY_SET", "STRING", "WS" ]

    RULE_constraint = 0
    RULE_expr = 1
    RULE_variable = 2
    RULE_elements = 3

    ruleNames =  [ "constraint", "expr", "variable", "elements" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    T__12=13
    T__13=14
    T__14=15
    T__15=16
    T__16=17
    T__17=18
    T__18=19
    T__19=20
    T__20=21
    T__21=22
    T__22=23
    T__23=24
    T__24=25
    T__25=26
    T__26=27
    T__27=28
    T__28=29
    T__29=30
    T__30=31
    T__31=32
    T__32=33
    BOOL=34
    VARIABLE=35
    INT=36
    FLOAT=37
    EMPTY_SET=38
    STRING=39
    WS=40

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.10.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class ConstraintContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expr(self):
            return self.getTypedRuleContext(winvalParser.ExprContext,0)


        def EOF(self):
            return self.getToken(winvalParser.EOF, 0)

        def getRuleIndex(self):
            return winvalParser.RULE_constraint

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterConstraint" ):
                listener.enterConstraint(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitConstraint" ):
                listener.exitConstraint(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitConstraint" ):
                return visitor.visitConstraint(self)
            else:
                return visitor.visitChildren(self)




    def constraint(self):

        localctx = winvalParser.ConstraintContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_constraint)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 8
            self.expr(0)
            self.state = 9
            self.match(winvalParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ExprContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return winvalParser.RULE_expr

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class AccessExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.struct = None # VariableContext
            self.key = None # ExprContext
            self.copyFrom(ctx)

        def variable(self):
            return self.getTypedRuleContext(winvalParser.VariableContext,0)

        def expr(self):
            return self.getTypedRuleContext(winvalParser.ExprContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAccessExpr" ):
                listener.enterAccessExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAccessExpr" ):
                listener.exitAccessExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAccessExpr" ):
                return visitor.visitAccessExpr(self)
            else:
                return visitor.visitChildren(self)


    class StringExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.atom = None # Token
            self.copyFrom(ctx)

        def STRING(self):
            return self.getToken(winvalParser.STRING, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStringExpr" ):
                listener.enterStringExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStringExpr" ):
                listener.exitStringExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitStringExpr" ):
                return visitor.visitStringExpr(self)
            else:
                return visitor.visitChildren(self)


    class BoolExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.atom = None # Token
            self.copyFrom(ctx)

        def BOOL(self):
            return self.getToken(winvalParser.BOOL, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBoolExpr" ):
                listener.enterBoolExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBoolExpr" ):
                listener.exitBoolExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitBoolExpr" ):
                return visitor.visitBoolExpr(self)
            else:
                return visitor.visitChildren(self)


    class FloatExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.atom = None # Token
            self.copyFrom(ctx)

        def FLOAT(self):
            return self.getToken(winvalParser.FLOAT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFloatExpr" ):
                listener.enterFloatExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFloatExpr" ):
                listener.exitFloatExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitFloatExpr" ):
                return visitor.visitFloatExpr(self)
            else:
                return visitor.visitChildren(self)


    class SetExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def elements(self):
            return self.getTypedRuleContext(winvalParser.ElementsContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSetExpr" ):
                listener.enterSetExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSetExpr" ):
                listener.exitSetExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSetExpr" ):
                return visitor.visitSetExpr(self)
            else:
                return visitor.visitChildren(self)


    class VariableExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def variable(self):
            return self.getTypedRuleContext(winvalParser.VariableContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVariableExpr" ):
                listener.enterVariableExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVariableExpr" ):
                listener.exitVariableExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitVariableExpr" ):
                return visitor.visitVariableExpr(self)
            else:
                return visitor.visitChildren(self)


    class IntExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.atom = None # Token
            self.copyFrom(ctx)

        def INT(self):
            return self.getToken(winvalParser.INT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterIntExpr" ):
                listener.enterIntExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitIntExpr" ):
                listener.exitIntExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitIntExpr" ):
                return visitor.visitIntExpr(self)
            else:
                return visitor.visitChildren(self)


    class ListExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def elements(self):
            return self.getTypedRuleContext(winvalParser.ElementsContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterListExpr" ):
                listener.enterListExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitListExpr" ):
                listener.exitListExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitListExpr" ):
                return visitor.visitListExpr(self)
            else:
                return visitor.visitChildren(self)


    class EmptySetExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.atom = None # Token
            self.copyFrom(ctx)

        def EMPTY_SET(self):
            return self.getToken(winvalParser.EMPTY_SET, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEmptySetExpr" ):
                listener.enterEmptySetExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEmptySetExpr" ):
                listener.exitEmptySetExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitEmptySetExpr" ):
                return visitor.visitEmptySetExpr(self)
            else:
                return visitor.visitChildren(self)


    class ParenExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(winvalParser.ExprContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterParenExpr" ):
                listener.enterParenExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitParenExpr" ):
                listener.exitParenExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitParenExpr" ):
                return visitor.visitParenExpr(self)
            else:
                return visitor.visitChildren(self)


    class InfixExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.left = None # ExprContext
            self.op = None # Token
            self.right = None # ExprContext
            self.copyFrom(ctx)

        def expr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(winvalParser.ExprContext)
            else:
                return self.getTypedRuleContext(winvalParser.ExprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInfixExpr" ):
                listener.enterInfixExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInfixExpr" ):
                listener.exitInfixExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitInfixExpr" ):
                return visitor.visitInfixExpr(self)
            else:
                return visitor.visitChildren(self)


    class PrefixExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.ExprContext
            super().__init__(parser)
            self.op = None # Token
            self.operand = None # ExprContext
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(winvalParser.ExprContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPrefixExpr" ):
                listener.enterPrefixExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPrefixExpr" ):
                listener.exitPrefixExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPrefixExpr" ):
                return visitor.visitPrefixExpr(self)
            else:
                return visitor.visitChildren(self)



    def expr(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = winvalParser.ExprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 2
        self.enterRecursionRule(localctx, 2, self.RULE_expr, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 37
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,0,self._ctx)
            if la_ == 1:
                localctx = winvalParser.IntExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 12
                localctx.atom = self.match(winvalParser.INT)
                pass

            elif la_ == 2:
                localctx = winvalParser.FloatExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 13
                localctx.atom = self.match(winvalParser.FLOAT)
                pass

            elif la_ == 3:
                localctx = winvalParser.StringExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 14
                localctx.atom = self.match(winvalParser.STRING)
                pass

            elif la_ == 4:
                localctx = winvalParser.BoolExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 15
                localctx.atom = self.match(winvalParser.BOOL)
                pass

            elif la_ == 5:
                localctx = winvalParser.EmptySetExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 16
                localctx.atom = self.match(winvalParser.EMPTY_SET)
                pass

            elif la_ == 6:
                localctx = winvalParser.VariableExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 17
                self.variable()
                pass

            elif la_ == 7:
                localctx = winvalParser.AccessExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 18
                localctx.struct = self.variable()
                self.state = 19
                self.match(winvalParser.T__0)
                self.state = 20
                localctx.key = self.expr(0)
                self.state = 21
                self.match(winvalParser.T__1)
                pass

            elif la_ == 8:
                localctx = winvalParser.ListExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 23
                self.match(winvalParser.T__0)
                self.state = 24
                self.elements()
                self.state = 25
                self.match(winvalParser.T__1)
                pass

            elif la_ == 9:
                localctx = winvalParser.SetExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 27
                self.match(winvalParser.T__2)
                self.state = 28
                self.elements()
                self.state = 29
                self.match(winvalParser.T__3)
                pass

            elif la_ == 10:
                localctx = winvalParser.ParenExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 31
                self.match(winvalParser.T__4)
                self.state = 32
                self.expr(0)
                self.state = 33
                self.match(winvalParser.T__5)
                pass

            elif la_ == 11:
                localctx = winvalParser.PrefixExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 35
                localctx.op = self._input.LT(1)
                _la = self._input.LA(1)
                if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << winvalParser.T__6) | (1 << winvalParser.T__7) | (1 << winvalParser.T__8) | (1 << winvalParser.T__9) | (1 << winvalParser.T__10) | (1 << winvalParser.T__11) | (1 << winvalParser.T__12))) != 0)):
                    localctx.op = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 36
                localctx.operand = self.expr(7)
                pass


            self._ctx.stop = self._input.LT(-1)
            self.state = 59
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,2,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 57
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
                    if la_ == 1:
                        localctx = winvalParser.InfixExprContext(self, winvalParser.ExprContext(self, _parentctx, _parentState))
                        localctx.left = _prevctx
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 39
                        if not self.precpred(self._ctx, 6):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")
                        self.state = 40
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << winvalParser.T__13) | (1 << winvalParser.T__14) | (1 << winvalParser.T__15) | (1 << winvalParser.T__16) | (1 << winvalParser.T__17) | (1 << winvalParser.T__18))) != 0)):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 41
                        localctx.right = self.expr(7)
                        pass

                    elif la_ == 2:
                        localctx = winvalParser.InfixExprContext(self, winvalParser.ExprContext(self, _parentctx, _parentState))
                        localctx.left = _prevctx
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 42
                        if not self.precpred(self._ctx, 5):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
                        self.state = 43
                        localctx.op = self.match(winvalParser.T__19)
                        self.state = 44
                        localctx.right = self.expr(6)
                        pass

                    elif la_ == 3:
                        localctx = winvalParser.InfixExprContext(self, winvalParser.ExprContext(self, _parentctx, _parentState))
                        localctx.left = _prevctx
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 45
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 4)")
                        self.state = 46
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not(_la==winvalParser.T__20 or _la==winvalParser.T__21):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 47
                        localctx.right = self.expr(5)
                        pass

                    elif la_ == 4:
                        localctx = winvalParser.InfixExprContext(self, winvalParser.ExprContext(self, _parentctx, _parentState))
                        localctx.left = _prevctx
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 48
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
                        self.state = 49
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << winvalParser.T__22) | (1 << winvalParser.T__23) | (1 << winvalParser.T__24) | (1 << winvalParser.T__25) | (1 << winvalParser.T__26) | (1 << winvalParser.T__27))) != 0)):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 50
                        localctx.right = self.expr(4)
                        pass

                    elif la_ == 5:
                        localctx = winvalParser.InfixExprContext(self, winvalParser.ExprContext(self, _parentctx, _parentState))
                        localctx.left = _prevctx
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 51
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                        self.state = 52
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not(_la==winvalParser.T__28 or _la==winvalParser.T__29):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 53
                        localctx.right = self.expr(3)
                        pass

                    elif la_ == 6:
                        localctx = winvalParser.InfixExprContext(self, winvalParser.ExprContext(self, _parentctx, _parentState))
                        localctx.left = _prevctx
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 54
                        if not self.precpred(self._ctx, 1):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 1)")
                        self.state = 55
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not(_la==winvalParser.T__30 or _la==winvalParser.T__31):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 56
                        localctx.right = self.expr(2)
                        pass

             
                self.state = 61
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,2,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx


    class VariableContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return winvalParser.RULE_variable

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class VariableExpr2Context(VariableContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a winvalParser.VariableContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def VARIABLE(self):
            return self.getToken(winvalParser.VARIABLE, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVariableExpr2" ):
                listener.enterVariableExpr2(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVariableExpr2" ):
                listener.exitVariableExpr2(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitVariableExpr2" ):
                return visitor.visitVariableExpr2(self)
            else:
                return visitor.visitChildren(self)



    def variable(self):

        localctx = winvalParser.VariableContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_variable)
        try:
            localctx = winvalParser.VariableExpr2Context(self, localctx)
            self.enterOuterAlt(localctx, 1)
            self.state = 62
            self.match(winvalParser.VARIABLE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ElementsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT(self, i:int=None):
            if i is None:
                return self.getTokens(winvalParser.INT)
            else:
                return self.getToken(winvalParser.INT, i)

        def FLOAT(self, i:int=None):
            if i is None:
                return self.getTokens(winvalParser.FLOAT)
            else:
                return self.getToken(winvalParser.FLOAT, i)

        def STRING(self, i:int=None):
            if i is None:
                return self.getTokens(winvalParser.STRING)
            else:
                return self.getToken(winvalParser.STRING, i)

        def VARIABLE(self, i:int=None):
            if i is None:
                return self.getTokens(winvalParser.VARIABLE)
            else:
                return self.getToken(winvalParser.VARIABLE, i)

        def getRuleIndex(self):
            return winvalParser.RULE_elements

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterElements" ):
                listener.enterElements(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitElements" ):
                listener.exitElements(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitElements" ):
                return visitor.visitElements(self)
            else:
                return visitor.visitChildren(self)




    def elements(self):

        localctx = winvalParser.ElementsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_elements)
        self._la = 0 # Token type
        try:
            self.state = 96
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [winvalParser.INT]:
                self.enterOuterAlt(localctx, 1)
                self.state = 64
                self.match(winvalParser.INT)
                self.state = 69
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==winvalParser.T__32:
                    self.state = 65
                    self.match(winvalParser.T__32)

                    self.state = 66
                    self.match(winvalParser.INT)
                    self.state = 71
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                pass
            elif token in [winvalParser.FLOAT]:
                self.enterOuterAlt(localctx, 2)
                self.state = 72
                self.match(winvalParser.FLOAT)
                self.state = 77
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==winvalParser.T__32:
                    self.state = 73
                    self.match(winvalParser.T__32)

                    self.state = 74
                    self.match(winvalParser.FLOAT)
                    self.state = 79
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                pass
            elif token in [winvalParser.STRING]:
                self.enterOuterAlt(localctx, 3)
                self.state = 80
                self.match(winvalParser.STRING)
                self.state = 85
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==winvalParser.T__32:
                    self.state = 81
                    self.match(winvalParser.T__32)

                    self.state = 82
                    self.match(winvalParser.STRING)
                    self.state = 87
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                pass
            elif token in [winvalParser.VARIABLE]:
                self.enterOuterAlt(localctx, 4)
                self.state = 88
                self.match(winvalParser.VARIABLE)
                self.state = 93
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==winvalParser.T__32:
                    self.state = 89
                    self.match(winvalParser.T__32)

                    self.state = 90
                    self.match(winvalParser.VARIABLE)
                    self.state = 95
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



    def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[1] = self.expr_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def expr_sempred(self, localctx:ExprContext, predIndex:int):
            if predIndex == 0:
                return self.precpred(self._ctx, 6)
         

            if predIndex == 1:
                return self.precpred(self._ctx, 5)
         

            if predIndex == 2:
                return self.precpred(self._ctx, 4)
         

            if predIndex == 3:
                return self.precpred(self._ctx, 3)
         

            if predIndex == 4:
                return self.precpred(self._ctx, 2)
         

            if predIndex == 5:
                return self.precpred(self._ctx, 1)
         







from __future__ import annotations

import argparse
import json
from os.path import basename, splitext

from antlr4 import *

from winval.antlr.winvalLexer import winvalLexer
from winval.antlr.winvalParser import winvalParser
from winval.antlr.winvalVisitor import winvalVisitor
from winval.python_extensions import iff, imply, defined, prefix, suffix
from winval.wdl_parser import WdlParser
from winval.workflow_var import WorkflowVar


def strip_string(string: str) -> str:
    return string.replace('"', '').replace("'", '')


class ConstraintEval(winvalVisitor):

    def __init__(self, workflow_variables: dict[str, WorkflowVar]):
        self.workflow_variables = workflow_variables

    def visitConstraint(self, ctx):
        if not ctx.EOF():
            # evaluate expression for further error reporting
            self.visit(ctx.expr())
            return False
        expression_value = self.visit(ctx.expr())
        if type(expression_value) != bool:
            raise ValueError(f'constraint evaluated to non-boolean value: {expression_value} , {ctx.getText()}')
        return expression_value

    def visitIntExpr(self, ctx):
        return int(ctx.getText())

    def visitBoolExpr(self, ctx):
        return ctx.getText() == 'True'

    def visitFloatExpr(self, ctx):
        return float(ctx.getText())

    def visitStringExpr(self, ctx):
        return strip_string(ctx.getText())

    def visitEmptySetExpr(self, ctx):
        return set()

    def visitAccessExpr(self, ctx):
        collection = self.visit(ctx.struct)
        key = self.visit(ctx.key)
        if type(collection) == dict:
            return collection[key] if key in collection else None
        else:
            return collection[key]

    def visitListExpr(self, ctx):
        if len(ctx.elements().INT()) > 0:
            return [int(elm.getText()) for elm in ctx.elements().INT()]
        elif len(ctx.elements().FLOAT()) > 0:
            return [float(elm.getText()) for elm in ctx.elements().FLOAT()]
        elif len(ctx.elements().STRING()) > 0:
            return [strip_string(elm.getText()) for elm in ctx.elements().STRING()]
        elif len(ctx.elements().VARIABLE()) > 0:
            return [self.workflow_variables[elm.getText()].value for elm in ctx.elements().VARIABLE()]

    def visitSetExpr(self, ctx):
        return set(self.visitListExpr(ctx))

    def visitParenExpr(self, ctx):
        return self.visit(ctx.expr())

    def visitInfixExpr(self, ctx):
        left = self.visit(ctx.left)
        op = ctx.op.text

        # Don't eval right expression (could contain runtime error) under some cases
        if op == '->' and left is False:
            return True
        if op == 'and' and left is False:
            return False
        if op == 'or' and left is True:
            return True

        right = self.visit(ctx.right)

        operation = {
            '<->': lambda: iff(left, right),
            '->': lambda: imply(left, right),
            'and': lambda: left and right,
            'or': lambda: left or right,
            '>': lambda: left > right,
            '>=': lambda: left >= right,
            '==': lambda: left == right,
            '!=': lambda: left != right,
            '<=': lambda: left <= right,
            '<': lambda: left < right,
            'in': lambda: left in right,
            '+': lambda: left + right,
            '-': lambda: left - right,
            '*': lambda: left * right,
            '/': lambda: left / right,
            '**': lambda: left ** right,
            '&': lambda: left & right,
            '|': lambda: left | right,
            '%': lambda: left % right,
        }
        return operation.get(op, lambda: None)()

    def visitPrefixExpr(self, ctx):
        operand = self.visit(ctx.operand)
        op = ctx.op.text
        operation = {
            'defined': lambda: defined(operand),
            'len': lambda: len(operand),
            'set': lambda: set(operand),
            'prefix': lambda: prefix(operand),
            'suffix': lambda: suffix(operand),
            'basename': lambda: basename(operand),
            'splitext': lambda: splitext(operand),
            'not': lambda: not operand,
        }
        return operation.get(op, lambda: None)()

    def visitVariableExpr(self, ctx):
        if ctx.getText() in self.workflow_variables:
            return self.workflow_variables[ctx.getText()].value
        else:
            raise ValueError(f'Undeclared variable: {ctx.getText()}\n')

    def visitVariableExpr2(self, ctx):
        return self.visitVariableExpr(ctx)


def get_args():
    parser = argparse.ArgumentParser("winval")
    parser.add_argument('--wdl', required=True)
    parser.add_argument('--json', required=True)
    parser.add_argument('--log_level', default='INFO')
    return parser.parse_args()


def main():
    args = get_args()
    workflow_inputs = WdlParser(args.wdl).parse_workflow_variables()
    with open(args.json) as json_file:
        json_dict = json.load(json_file)
        workflow_inputs.fill_values_from_json(json_dict)

    while 1:
        data = InputStream(input(">>> "))
        # lexer
        lexer = winvalLexer(data)
        stream = CommonTokenStream(lexer)

        # parser
        parser = winvalParser(stream)
        tree = parser.constraint()
        # evaluator
        visitor = ConstraintEval(workflow_inputs.workflow_vars)
        try:
            output = visitor.visit(tree)
            print(output)
        except ValueError as e:
            print(e)


if __name__ == "__main__":
    main()



import logging
import sys
from os.path import dirname

base_dir = dirname(__file__)

# create logger
logger = logging.getLogger('winval')
logger.setLevel(logging.DEBUG)

# create formatter
formatter = logging.Formatter('%(asctime)s - %(module)s - %(levelname)s - %(message)s')

# create console handler and set level to debug
ch = logging.StreamHandler(stream=sys.stdout)
ch.setLevel(logging.INFO)
ch.setFormatter(formatter)
logger.addHandler(ch)

# add package paths to PYTHONPATH
path = f'{dirname(dirname(__file__))}'
if path not in sys.path:
    sys.path.append(path)



from enum import Enum, auto
from winval import logger
from typing import Optional


class WorkflowVarType(Enum):
    STRING = auto()
    INT = auto()
    FLOAT = auto()
    FILE = auto()
    BOOLEAN = auto()
    STRING_ARRAY = auto()
    INT_ARRAY = auto()
    FLOAT_ARRAY = auto()
    FILE_ARRAY = auto()
    BOOLEAN_ARRAY = auto()
    FILE_MATRIX = auto()
    FLOAT_MATRIX = auto()
    INT_MATRIX = auto()
    STRING_MATRIX = auto()
    FLOAT_PAIRS = auto()
    INT_PAIRS = auto()
    STRUCT = auto()
    MAP = auto()
    PAIR = auto()

    @staticmethod
    def from_string(string: str):
        if string == 'String':
            return WorkflowVarType.STRING
        elif string == 'Int':
            return WorkflowVarType.INT
        elif string == 'Float':
            return WorkflowVarType.FLOAT
        elif string == 'File':
            return WorkflowVarType.FILE
        elif string == 'Boolean':
            return WorkflowVarType.BOOLEAN
        elif string == 'Array[String]':
            return WorkflowVarType.STRING_ARRAY
        elif string == 'Array[Int]':
            return WorkflowVarType.INT_ARRAY
        elif string == 'Array[Float]':
            return WorkflowVarType.FLOAT_ARRAY
        elif string == 'Array[File]':
            return WorkflowVarType.FILE_ARRAY
        elif string == 'Array[Boolean]':
            return WorkflowVarType.BOOLEAN_ARRAY
        elif string == 'Array[Array[File]]':
            return WorkflowVarType.FILE_MATRIX
        elif string == 'Array[Array[Float]]':
            return WorkflowVarType.FLOAT_MATRIX
        elif string == 'Array[Array[Int]]':
            return WorkflowVarType.INT_MATRIX
        elif string == 'Array[Array[String]]':
            return WorkflowVarType.STRING_MATRIX
        elif string == 'Array[Pair[Float]]':
            return WorkflowVarType.FLOAT_PAIRS
        elif string == 'Array[Pair[Float]]':
            return WorkflowVarType.INT_PAIRS
        elif string.startswith('Map'):
            return WorkflowVarType.MAP
        elif string.startswith('Pair'):
            return WorkflowVarType.PAIR
        else:
            raise ValueError(f'No such WorkflowVarType: {string}')


class WorkflowVar:

    def __init__(self, name: str,
                 var_type: WorkflowVarType,
                 is_optional: bool,
                 default_value: Optional[str],
                 workflow_name: str,
                 struct_id: str = None):
        self.name = name
        self.type = var_type
        self.is_optional = is_optional
        self.value = self.__parse_default_value_from_str(default_value)
        self.workflow_name = workflow_name
        self.full_name = f'{workflow_name}.{name}'
        self.struct_id = struct_id

    def __repr__(self):
        is_opt = '?' if self.is_optional else ''
        return f'{self.name} {self.type} {is_opt} {self.value}'

    def set_value(self, value):
        self.value = value

    def __parse_default_value_from_str(self, value: str):
        if value is None:
            return None
        if self.type == WorkflowVarType.INT:
            try:
                return int(value)
            except ValueError:
                logger.warning(f'Ignore parsing complex default int expression {self.type} {self.name} {value}')
                return None
        if self.type == WorkflowVarType.FLOAT:
            try:
                return float(value)
            except ValueError:
                logger.warning(f'Ignore parsing complex default int expression {self.type} {self.name} {value}')
                return None

        if self.type == WorkflowVarType.BOOLEAN:
            if value == 'true':
                return True
            elif value == "false":
                return False
            else:
                logger.warning(f'Ignore parsing complex default bool expression {self.type} {self.name} {value}')
                return None
        if self.type in {WorkflowVarType.STRING_ARRAY,
                         WorkflowVarType.FILE_ARRAY,
                         WorkflowVarType.INT_ARRAY,
                         WorkflowVarType.BOOLEAN_ARRAY,
                         WorkflowVarType.FLOAT_ARRAY}:
            if value[0] not in {'['}:
                logger.warning(f'Ignore parsing complex default array value {self.type} {self.name} {value}')
                return None
            if value == '[]':
                return []
            else:
                str_array = value.strip('][ ').split(',')
                if self.type == WorkflowVarType.INT_ARRAY:
                    return [int(x) for x in str_array]
                if self.type == WorkflowVarType.FLOAT_ARRAY:
                    return [float(x) for x in str_array]
                if self.type == WorkflowVarType.BOOLEAN_ARRAY:
                    return [x == 'true' for x in str_array]
                else:
                    return [x.strip().strip('"').strip("'") for x in str_array]
        if self.type in {WorkflowVarType.MAP, WorkflowVarType.PAIR}:
            logger.warning(f'Ignore parsing default of Map and Pair types: {self.type} {self.name} {value}')
            return None
        return value



# Copyright 2022 Ultima Genomics Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# DESCRIPTION
#    Validate all files defined by  workflow inputs json conf are found in google-cloud-storage

import argparse
import json
import re
import google
import subprocess
from google.cloud import storage
import boto3
import botocore
from enum import Enum
from winval.workflow_var import WorkflowVarType
from winval.wdl_parser import WdlParser
from winval import logger


STORAGE_PREFIX_REGEX = "^^(gs|s3)://"

#StorageClass could be EXIST,MISSING,GLACIER

class StorageClass(Enum):
    EXIST = 1
    MISSING = 2
    GLACIER = 3

def get_args():
    parser = argparse.ArgumentParser('cloud_files_validator', description='Validate GCS inputs actually exist')
    parser.add_argument('--json', required=True)
    parser.add_argument('--wdl', required=True)
    return parser.parse_args()


def split_uri(uri: str):
    matches = re.match(rf"{STORAGE_PREFIX_REGEX}(.*?)/(.*)", uri)
    if matches:
        return matches.groups()
    else:
        return None, None


class CloudFilesValidator:

    def __init__(self, wdl_file: str, json_file: str):
        logger.debug('-----------------------------')
        logger.debug('---- CloudFilesValidator ----')
        logger.debug('-----------------------------')
        self.json_file = json_file
        self.workflow_vars = None
        wdl_inputs = WdlParser(wdl_file).parse_workflow_variables()
        self.workflow_vars = wdl_inputs.workflow_vars
        with open(json_file) as jf:
            wdl_inputs.fill_values_from_json(json.load(jf))
        self.validated_files = []
        self.non_validated_files = []
        self.glacier_files = []


    def validate(self) -> bool:
        for workflow_var in self.workflow_vars.values():
            if workflow_var.type == WorkflowVarType.FILE and workflow_var.value is not None:
                self.validate_file(workflow_var.value)
            elif workflow_var.type == WorkflowVarType.FILE_ARRAY and workflow_var.value is not None:
                for file_uri in workflow_var.value:
                    self.validate_file(file_uri)

        logger.debug('-------------------------------------------')
        summary_str = f"Existing URI's: {len(self.validated_files)}, Non-existing URI's: {len(self.non_validated_files)}, Glacier URI's: {len(self.glacier_files)}"
        logger.info(summary_str)
        for fn in self.non_validated_files:
            logger.error(f'Missing URI: {fn}')
        for fn in self.glacier_files:
            logger.error(f'Glacier URI: {fn}')

        return len(self.non_validated_files) == 0


    def validate_file(self, file_uri: str):
        file_uri = file_uri.replace('"', '')
        if not re.match(fr"{STORAGE_PREFIX_REGEX}", file_uri):
            return
        try:
            proto, bucket_name, file_name = split_uri(file_uri)
        except AttributeError:
            self.non_validated_files.append(file_uri)
        else:
            exists = self.blob_exists(bucket_name, file_name, storage_type=proto)
            if exists == StorageClass.EXIST:
                self.validated_files.append(file_uri)
            else:
                self.non_validated_files.append(file_uri)
            if exists == StorageClass.GLACIER:
                self.glacier_file(file_uri).append(file_uri)

    def blob_exists(self, bucket_name, file_name, storage_type="gs"):
        exists = StorageClass.MISSING
        if storage_type == "gs":
            try:
                client = storage.Client()
                bucket = client.get_bucket(bucket_name)
                blob = bucket.blob(file_name)
                exists = blob.exists()
            # blob_exists requires storage.bucket.get access which might be forbidden, 
            # yet objects in this bucket can still be accessed by gsutil
            # Could not fine a python api to access the object in this case
            except google.api_core.exceptions.Forbidden:
                full_uri = f"gs://{bucket_name}/{file_name}"
                process = subprocess.run(f'gsutil ls {full_uri}'.split(), capture_output=True)
                if process.returncode == 0:
                    exists = StorageClass.EXIST
        if storage_type == "s3":
            try:
                client = boto3.client("s3")
                ho = client.head_object(
                    Bucket=bucket_name,
                    Key=file_name
                )
                if 'StorageClass' in ho and ho['StorageClass'] == 'GLACIER':
                    return StorageClass.GLACIER
                exists = StorageClass.EXIST
            except botocore.exceptions.ClientError:
                pass
        return exists
                

    def validate_files_in_object(self, o):
        if type(o) is dict:
            for key, value in o.items():
                self.validate_files_in_object(value)
        elif type(o) is list:
            for value in o:
                self.validate_files_in_object(value)
        elif type(o) is str:
            self.validate_file(o)


def main():
    """
    validate all GCP/AWS URLs in json inputs point to existing objects
    """
    args = get_args()
    validated = CloudFilesValidator(args.wdl, args.json).validate()
    if not validated:
        raise RuntimeError('Found non existing files!')


if __name__ == '__main__':
    main()



import argparse
import json

# Copyright 2022 Ultima Genomics Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# DESCRIPTION
#    Validate set of constraints defined on workflow inputs, as defined inside a wdl file, on a specified input json


from winval.wdl_parser import WdlParser
from winval.constraints_pre_processor import ConstraintsPreProcessor
from winval.winval_class import Winval
from winval import logger


def run_winval(wdl_file: str, json_file: str) -> bool:
    logger.debug('--------------')
    logger.debug('--- WINVAL ---')
    logger.debug('--------------')
    with open(json_file) as json_file:
        json_dict = json.load(json_file)
        workflow_inputs = WdlParser(wdl_file).parse_workflow_variables()
        workflow_inputs.fill_values_from_json(json_dict)
        constraints = ConstraintsPreProcessor(wdl_file).process_constraint_strings()
        return Winval(workflow_inputs, constraints).workflow_input_validation()
    

def get_args():
    parser = argparse.ArgumentParser("winval")
    parser.add_argument('--wdl', required=True)
    parser.add_argument('--json', required=True)
    parser.add_argument('--log_level', default='INFO')
    return parser.parse_args()


if __name__ == '__main__':
    args = get_args()
    [handler.setLevel(args.log_level.upper()) for handler in logger.handlers]
    run_winval(args.wdl, args.json)



# Generated from winval.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .winvalParser import winvalParser
else:
    from winvalParser import winvalParser

# This class defines a complete generic visitor for a parse tree produced by winvalParser.

class winvalVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by winvalParser#constraint.
    def visitConstraint(self, ctx:winvalParser.ConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#AccessExpr.
    def visitAccessExpr(self, ctx:winvalParser.AccessExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#StringExpr.
    def visitStringExpr(self, ctx:winvalParser.StringExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#BoolExpr.
    def visitBoolExpr(self, ctx:winvalParser.BoolExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#FloatExpr.
    def visitFloatExpr(self, ctx:winvalParser.FloatExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#SetExpr.
    def visitSetExpr(self, ctx:winvalParser.SetExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#VariableExpr.
    def visitVariableExpr(self, ctx:winvalParser.VariableExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#IntExpr.
    def visitIntExpr(self, ctx:winvalParser.IntExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#ListExpr.
    def visitListExpr(self, ctx:winvalParser.ListExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#EmptySetExpr.
    def visitEmptySetExpr(self, ctx:winvalParser.EmptySetExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#ParenExpr.
    def visitParenExpr(self, ctx:winvalParser.ParenExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#InfixExpr.
    def visitInfixExpr(self, ctx:winvalParser.InfixExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#PrefixExpr.
    def visitPrefixExpr(self, ctx:winvalParser.PrefixExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#VariableExpr2.
    def visitVariableExpr2(self, ctx:winvalParser.VariableExpr2Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by winvalParser#elements.
    def visitElements(self, ctx:winvalParser.ElementsContext):
        return self.visitChildren(ctx)



del winvalParser


from __future__ import annotations
from winval.workflow_var import WorkflowVar, WorkflowVarType


class ParsedWorkflowInputs:
    """
    Stores workflow vars and name-space structs
    Fills values of variables from json inputs configuration
    """

    def __init__(self,
                 workflow_vars: dict[str, WorkflowVar],
                 structs: dict[str, dict[str, WorkflowVar]]):
        self.workflow_vars = workflow_vars
        self.structs = structs

    def fill_values_from_json(self, json_dict: dict):
        for var in self.workflow_vars.values():
            if var.full_name in json_dict:
                var.set_value(json_dict[var.full_name])
                if var.type == WorkflowVarType.STRUCT:
                    self.__validate_struct_assignment(var)
            else:
                if not var.is_optional and var.value is None:
                    raise ValueError(f'Mandatory input {var} is not set')

        for json_key in json_dict.keys():
            simple_name = json_key.split('.')[-1]
            if simple_name not in self.workflow_vars:
                raise ValueError(f'Variable {json_key} is not defined in workflow')

    def __validate_struct_assignment(self, struct_variable: WorkflowVar):
        struct_fields = self.structs[struct_variable.struct_id]
        field_values_dict = struct_variable.value
        for struct_field in struct_fields.values():
            if not struct_field.is_optional and struct_field.name not in field_values_dict:
                raise ValueError(f'Struct field {struct_field} is not set for used struct {struct_variable}')
        for json_struct_field in field_values_dict.keys():
            if json_struct_field not in struct_fields:
                raise ValueError(f'Field {json_struct_field} is not defined for struct {struct_variable.struct_id}')




from __future__ import annotations

from os.path import basename

from antlr4 import CommonTokenStream, InputStream
from antlr4.error.ErrorListener import ErrorListener

from winval.antlr.winvalLexer import winvalLexer
from winval.antlr.winvalParser import winvalParser
from winval.parsed_workflow_inputs import ParsedWorkflowInputs
from winval.python_extensions import defined, prefix, suffix, iff, imply
from winval.constraint_eval import ConstraintEval
from winval import logger


class MyErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offending_symbol, line, column, msg, e):
        raise SyntaxError("ERROR: when parsing line %d column %d: %s\n" % (line, column, msg))


class Winval:

    def __init__(self, workflow_inputs: ParsedWorkflowInputs, constraints: list[str]):
        self.workflow_inputs = workflow_inputs
        self.constraints = constraints
        self.__exemplify_extensions()

    def workflow_input_validation(self) -> bool:
        validations: list[tuple[bool, str]] = self.eval_constraints()
        passed = [v for v in validations if v[0]]
        failed = [v for v in validations if not v[0]]
        all_passed = len(failed) == 0
        if len(passed) > 0:
            logger.debug('Passed validations:')
        for validation in passed:
            logger.debug(f'True: {validation[1]}')
        if not all_passed:
            logger.error('Failed validations:')
        for validation in failed:
            logger.error(f'False: {validation[1]}')
        if all_passed:
            logger.info(f'Passed all {len(passed)} winval validations!')
        return all_passed

    def eval_constraints(self) -> list[tuple[bool, str]]:
        validations = []
        error_listener = MyErrorListener()
        for constraint_str in self.constraints:
            # lexer
            lexer = winvalLexer(InputStream(constraint_str))
            # def recover(self, e): pass
            # lexer.recover = recover
            lexer.removeErrorListeners()
            lexer.addErrorListener(error_listener)
            try:
                stream = CommonTokenStream(lexer)
                # parser
                parser = winvalParser(stream)
                tree = parser.constraint()
                # evaluator
                visitor = ConstraintEval(self.workflow_inputs.workflow_vars)
                try:
                    validation_result = visitor.visit(tree)
                except (KeyError, TypeError) as e:
                    validation_result = False
                    constraint_str += f' : ( {str(e).strip()} )'
            except (SyntaxError, ValueError) as e:
                validation_result = False
                constraint_str += f' : ( {str(e).strip()} )'
            validations.append((validation_result, constraint_str))
        return validations

    @staticmethod
    def __exemplify_extensions():
        # Use winval_python_extensions, so that import is not unused
        assert defined(6)
        assert prefix('a.bam') == 'a'
        assert suffix('a.bam') == '.bam'
        assert suffix(['a.bam']) == {'.bam'}
        assert basename('gs://bucket/a.bam') == 'a.bam'
        assert iff(True, True)
        assert imply(False, True)
        assert imply(1 > 2, True)
        assert imply(4 > 1, 2 > 1)



# Generated from winval.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .winvalParser import winvalParser
else:
    from winvalParser import winvalParser

# This class defines a complete listener for a parse tree produced by winvalParser.
class winvalListener(ParseTreeListener):

    # Enter a parse tree produced by winvalParser#constraint.
    def enterConstraint(self, ctx:winvalParser.ConstraintContext):
        pass

    # Exit a parse tree produced by winvalParser#constraint.
    def exitConstraint(self, ctx:winvalParser.ConstraintContext):
        pass


    # Enter a parse tree produced by winvalParser#AccessExpr.
    def enterAccessExpr(self, ctx:winvalParser.AccessExprContext):
        pass

    # Exit a parse tree produced by winvalParser#AccessExpr.
    def exitAccessExpr(self, ctx:winvalParser.AccessExprContext):
        pass


    # Enter a parse tree produced by winvalParser#StringExpr.
    def enterStringExpr(self, ctx:winvalParser.StringExprContext):
        pass

    # Exit a parse tree produced by winvalParser#StringExpr.
    def exitStringExpr(self, ctx:winvalParser.StringExprContext):
        pass


    # Enter a parse tree produced by winvalParser#BoolExpr.
    def enterBoolExpr(self, ctx:winvalParser.BoolExprContext):
        pass

    # Exit a parse tree produced by winvalParser#BoolExpr.
    def exitBoolExpr(self, ctx:winvalParser.BoolExprContext):
        pass


    # Enter a parse tree produced by winvalParser#FloatExpr.
    def enterFloatExpr(self, ctx:winvalParser.FloatExprContext):
        pass

    # Exit a parse tree produced by winvalParser#FloatExpr.
    def exitFloatExpr(self, ctx:winvalParser.FloatExprContext):
        pass


    # Enter a parse tree produced by winvalParser#SetExpr.
    def enterSetExpr(self, ctx:winvalParser.SetExprContext):
        pass

    # Exit a parse tree produced by winvalParser#SetExpr.
    def exitSetExpr(self, ctx:winvalParser.SetExprContext):
        pass


    # Enter a parse tree produced by winvalParser#VariableExpr.
    def enterVariableExpr(self, ctx:winvalParser.VariableExprContext):
        pass

    # Exit a parse tree produced by winvalParser#VariableExpr.
    def exitVariableExpr(self, ctx:winvalParser.VariableExprContext):
        pass


    # Enter a parse tree produced by winvalParser#IntExpr.
    def enterIntExpr(self, ctx:winvalParser.IntExprContext):
        pass

    # Exit a parse tree produced by winvalParser#IntExpr.
    def exitIntExpr(self, ctx:winvalParser.IntExprContext):
        pass


    # Enter a parse tree produced by winvalParser#ListExpr.
    def enterListExpr(self, ctx:winvalParser.ListExprContext):
        pass

    # Exit a parse tree produced by winvalParser#ListExpr.
    def exitListExpr(self, ctx:winvalParser.ListExprContext):
        pass


    # Enter a parse tree produced by winvalParser#EmptySetExpr.
    def enterEmptySetExpr(self, ctx:winvalParser.EmptySetExprContext):
        pass

    # Exit a parse tree produced by winvalParser#EmptySetExpr.
    def exitEmptySetExpr(self, ctx:winvalParser.EmptySetExprContext):
        pass


    # Enter a parse tree produced by winvalParser#ParenExpr.
    def enterParenExpr(self, ctx:winvalParser.ParenExprContext):
        pass

    # Exit a parse tree produced by winvalParser#ParenExpr.
    def exitParenExpr(self, ctx:winvalParser.ParenExprContext):
        pass


    # Enter a parse tree produced by winvalParser#InfixExpr.
    def enterInfixExpr(self, ctx:winvalParser.InfixExprContext):
        pass

    # Exit a parse tree produced by winvalParser#InfixExpr.
    def exitInfixExpr(self, ctx:winvalParser.InfixExprContext):
        pass


    # Enter a parse tree produced by winvalParser#PrefixExpr.
    def enterPrefixExpr(self, ctx:winvalParser.PrefixExprContext):
        pass

    # Exit a parse tree produced by winvalParser#PrefixExpr.
    def exitPrefixExpr(self, ctx:winvalParser.PrefixExprContext):
        pass


    # Enter a parse tree produced by winvalParser#VariableExpr2.
    def enterVariableExpr2(self, ctx:winvalParser.VariableExpr2Context):
        pass

    # Exit a parse tree produced by winvalParser#VariableExpr2.
    def exitVariableExpr2(self, ctx:winvalParser.VariableExpr2Context):
        pass


    # Enter a parse tree produced by winvalParser#elements.
    def enterElements(self, ctx:winvalParser.ElementsContext):
        pass

    # Exit a parse tree produced by winvalParser#elements.
    def exitElements(self, ctx:winvalParser.ElementsContext):
        pass



del winvalParser


from __future__ import annotations


class ConstraintsPreProcessor:
    """
    Pre-process constraint strings before parsing
    """

    WINVAL_ANNOTATION = '#@wv'

    def __init__(self, wdl_file: str):
        self.wdl_file = wdl_file

    def process_constraint_strings(self, ) -> list[str]:
        with open(self.wdl_file) as wf:
            lines = wf.readlines()
            return self._process_constraint_strings_from_lines(lines)

    @staticmethod
    def _process_constraint_strings_from_lines(wdl_lines: list[str]) -> list[str]:
        constraints: list[str] = []
        for line in wdl_lines:
            line = line.strip()
            if line.startswith(ConstraintsPreProcessor.WINVAL_ANNOTATION):
                constraints.append(line.replace(ConstraintsPreProcessor.WINVAL_ANNOTATION, '').strip())
        return constraints



from os.path import splitext

"""
Define additional functions / operators that can be used in winval constraints syntax
winval syntax is used inside WDL #@wd comments
For example:

#@wv suffix(bam_index_files) == {".bai"}
"""


def iff(x, y): return (x and y) or not (x or y)
def imply(x, y): return not x or y
def defined(x): return x is not None


def prefix(x):
    if type(x) == str:
        return splitext(x)[0]
    elif type(x) == list:
        return [prefix(elm) for elm in x]
    else:
        raise ValueError(f'Can not prefix non-strings: {x} of type {type(x)}')


def suffix(x):
    if type(x) == str:
        return splitext(x)[1]
    elif type(x) == list:
        result = set()
        for elm in x:
            suf = suffix(elm)
            if type(suf) == str:
                result.add(suf)
            elif type(suf) == set:
                result.update(suf)
            else:
                raise ValueError(f"Can not add recursive suffix result {suf} to suffix set")

        return result
    else:
        raise ValueError("Can only suffix str/list[str]/list[list[str]]")
